/**
 *
 * $Id: j2b2GIMICtrl.hpp,v 1.28 2007-12-04 18:48:46 amaula Exp $
 *
 */

#ifndef _J2B2GIMICTRL_HPP_
#define _J2B2GIMICTRL_HPP_

#include <string>
#include <vector>
#include "ASRoboClient.hpp"
#include "CImageClient.h"

// Version of j2b2GIMICtrl
#define J2B2GIMICTRL_LIBRARY_VERSION_STRING "j2b2GIMICtrl 0.9.9-BETA (Build 16)"

/// Forward declaration of GIMI
namespace gimi {
  class GIMI;
}


// Predefined pins for the ASRobo_IO interface
#define ASROBO_IOPIN_SOLENOID    25
#define ASROBO_IOPIN_LIGHT       30
#define ASROBO_IOPIN_LASER       31


namespace J2B2GIMICtrl {

  class CJ2B2GIMICtrl {
  public:

    /**
     * Default constructor.
     *
     * @param[in] aGIMInt       If there is an existing GIMI interface available,
     *                          specify its pointer here. Otherwise the interface
     *                          instance will be created automatically.
     */
    CJ2B2GIMICtrl(gimi::GIMI *aGIMInt = NULL);


    /**
     * Destructor. Takes care of cleaning up the
     */
    virtual ~CJ2B2GIMICtrl();


    /**
     * This function returns the version details for this library revision as a single
     * string.
     *
     * @return                  Returns the current library version as a string.
     *
     */
    static std::string GetLibraryVersionStr(void) {
      return std::string(J2B2GIMICTRL_LIBRARY_VERSION_STRING);
    }


    /**
     * Connect to GIMnet. This is required to be able to do anything.
     * If this function fails, the j2b2GIMICtrl is unuseful.
     *
     * @param[in] aHubHost      Hostname of Hub to connect
     * @param[in] aHubPort      Port of Hub to connect
     * @param[in] aHubOwnname   Own name on the hub
     *
     * @return                  true = Connection succesfull
     *                          false = Connection failed
     */
    bool Connect(const std::string &aHubHost,
		 const int aHubPort,
		 const std::string &aHubOwnname);


    /**
     * This will set the known static delay of the link to 'aDelayMs' milliseconds.
     * This value means the time it takes to transmit a command over the communication
     * infrastructure to the receiving client. Setting this value to nonzero value
     * just increases the allowed timeouts on the client libraries.
     * 
     * @param[in] aDelayMs      Static delay of the link per one direction 
     *                          in milliseconds.
     * @return                  true = Value set successfully
     *                          false = Setting value failed.
     *
     */
    bool SetLinkStaticDelay(const unsigned int aDelayMs);


    /**
     * Initialize ASRobo (Robot Control). Attempts to connect
     * ASRobo Server, and initializes preconfigured services.
     * If any of the initializations fails, this call will return
     * false.
     *
     * @param[in] aServiceName  HubName of the service.
     * @return                  true = Initialization OK,
     *                            false = init failed!
     */
    bool InitializeASRobo(const std::string &aServiceName);


    /**
     * Initialize PTU (Pan-Tilt Unit). Attempts to connect to
     * PTU-Server. If the connection fails, this call will return
     * false.
     *
     * @param[in] aServiceName  HubName of the service.
     * @return                  true = Initialization OK,
     *                          false = init failed!
     */
    bool InitializePTU(const std::string &aServiceName);


    /**
     * Initialize ImageClient (Image Server Client). Attempts to
     * connect ImageServer. If connection fails, this call will
     * return false.
     *
     * @param[in] aServiceName  HubName of the service.
     * @return                  true = Initialization OK,
     *                          false = init failed!
     */
    bool InitializeImageClient(const std::string &aServiceName);



    /**************************************************************************
     *************** PTU Interface ********************************************
     *************************************************************************/

    /**
     * Turns the PTU unit to 'pan angle', 'tilt angle'.
     *
     * @param[in] aPan          Pan angle to turn in rads. Zero is centered,
     *                          positive angle is CCW, negative is CW.
     * @param[in] aTilt         Tilt angle to turn in rads. Zero is centered,
     *                          positive angle is 'up', negative angle is 'down'
     * @return                  true = Command succeeded
     *                          false = command failed!
     */
    bool PTU_TurnTo(const double aPan, const double aTilt);



    /**************************************************************************
     *************** RC Interface *********************************************
     *************************************************************************/

    /**
     * Sends data to RC-servo controller. The indexes 0 and 1 are used by the PTU. Rest
     * can be used for own applications
     *
     * @param[in] aServo        Servo channel (0-7)
     * @param[in] aPosition     RC-servo pulse "width" 0...255
     * @return                  true = Command succeeded
     *                          false = command failed!
     */
    bool RC_SetPosition(const unsigned char aServo, const unsigned char aPosition);



    /**************************************************************************
     *************** ASRobo Interface *****************************************
     *************************************************************************/

    /**
     * Set the robot Speed. All speed commands follow the current acceleration
     * value set. \See ASRobo_SetAcceleration. The speeds are limited per platform.
     *
     * For J2B2, the Maximum speed per wheel is 0.5m/s. This also dictates the
     * maximum angular speed, which for J2B2 is 3.84 rad/s.
     *
     *
     * @param[in] aSpeed        Base speed (m/s) Positive values move forward,
     *                          negative values backward.
     * @param[in] aAngularSpeed Angular speed (turning rate) (rad/s). Positive
     *                          values turn CCW, negative CW.
     * @return                  true = Command succeeded
     *                          false = command failed!
     */
    bool ASRobo_SetSpeed(const double aSpeed, const double aAngularSpeed);


    /**
     * Set the robot acceleration to 'aAcceleration'. The value of acceleration
     * affects all future 'SetSpeed' commands.
     *
     * For J2B2, the maximum possible acceleration is 0.2m/s�
     *
     * @param[in] aAcceleration Acceleration to set in 'm/s'
     * @return                  true = Command succeeded
     *                          false = command failed!
     */
    bool ASRobo_SetAcceleration(const double aAcceleration);


    /**
     * Stop the robot as fast as possible. (Faster than
     * setting robot speed to 0.0)
     *
     * @return                  true = Command succeeded
     *                          false = command failed!
     */
    bool ASRobo_Stop(bool reserved = false);


    /**
     * Sets the state of all available Behaviours to 'state'.
     *
     * For FSR07, the only behaviour enabled is the STOP behaviour,
     * So it is safe to call this function to Disable/Enable this feature.
     *
     *
     * @param[in] state         Status to set all behavioours to.
     * @return                  true = Command succeeded
     *                          false = command failed!
     */
    bool ASRobo_SetAllBehaviours(bool state);


    /**
     * Sets a coordinate to coordinateDrive-behaviour
     * @return                  true = Command succeeded
     *                          false = command failed!
     */
    bool ASRobo_SetCoordinate(float aX, float aY);


    /**
     * Sets the state of a specific behaviour by index.
     *
     * @param[in] aBehaviourIndex Index of behaviour to set
     * @param[in] aTurnOn       State to set (true == On, false == Off)
     * @return                  true = Command succeeded
     *                          false = command failed!
     */
    bool ASRobo_SetBehaviourByIndex(bool aTurnOn,int aBehaviourIndex);


    /**
     * Query for name of behaviour.
     *
     * @param[in] aIndex        Index to query name for
     * @return                  Name of the behaviour, or empty
     *                          string on error.
     */
    std::string ASRobo_GetNameOfBehaviour(int aIndex);


    /**
     * Returns the number of behaviours available.
     *
     * @return                  Number of behaviours available,
     *                          or negative on error.
     */
    int ASRobo_GetNumberOfBehaviours(void);


    /**
     * Returns a LaserScan result.
     *
     * @param[out] aDistanceArray     Data from the Laser Device. (reference is filled)
     * Format:
     *  double beam_spacing;          // Space between beams (rads)
     *  double beam_width;            // Width of beam in scan
     *	unsigned int timestamp_s;     // Timestamp in seconds
     *  unsigned int timestamp_us;    // Microseconds over the seconds
     *  TDistanceArray distanceArray; // Array of distances (see below)
     *
     * Where distanceArray is a STL Vector of CDistance:s (You can get the number
     * of measures available using the operator size() for the distanceArray member)
     *  double heading;               // Heading (rad), from relativePos, CCW
     *	double distance;              // Distance of target (m)
     *
     * @param[out] aSequence          Pointer to sequence number used for synchronized
     *                                waiting of new scans.
     *                                - If this parameter is NULL, the most recent
     *                                scan available is returned immediately.
     *                                - If this parameter is not-null, it should
     *                                point to integer having an initial value
     *                                of '-1'. After an succesfull call, the
     *                                value will be modified by the function to
     *                                contain the sequence of the most recent
     *                                scan read. After the call, use the same
     *                                pointer from call to call to obtain
     *                                all scans as fast as they come available.
     *
     * @return                        true = Command succeeded
     *                                false = command failed!
     */
    bool ASRobo_GetLaserScan(ASRobo::Drivers::Devices::Ranging::CDistanceArray &aDistanceArray,
			     int *aSequence = NULL);


    /**
     * Returns a BumperScan result. The format of data and call is identical
     * to LaserScan, because both devices are really the same type. Bumper devices
     * return negative distances for angles NOT HIT, and positive distances
     * for angles HIT. (Positive distances also reflect the distance of the bumper
     * from the center of robot)
     *
     * @param[out] aDistanceArray Data from the Device. (reference is filled)
     * Format:
     *  double beam_spacing;          // Space between beams (rads)
     *  double beam_width;            // Width of beam in scan
     *	unsigned int timestamp_s;     // Timestamp in seconds
     *  unsigned int timestamp_us;    // Microseconds over the seconds
     *  TDistanceArray distanceArray; // Array of distances (see below)
     *
     * Where distanceArray is a STL Vector of CDistance:s (You can get the number
     * of measures available using the operator size() for the distanceArray member)
     *  double heading;               // Heading (rad), from relativePos, CCW
     *	double distance;              // Distance of target (m)
     *
     * @param[out] aSequence          Pointer to sequence number used for synchronized
     *                                waiting of new scans.
     *                                - If this parameter is NULL, the most recent
     *                                scan available is returned immediately.
     *                                - If this parameter is not-null, it should
     *                                point to integer having an initial value
     *                                of '-1'. After an succesfull call, the
     *                                value will be modified by the function to
     *                                contain the sequence of the most recent
     *                                scan read. After the call, use the same
     *                                pointer from call to call to obtain
     *                                all scans as fast as they come available.
     *
     * @return                        true = Command succeeded
     *                                false = command failed!
     */
    bool ASRobo_GetBumperScan(ASRobo::Drivers::Devices::Ranging::CDistanceArray &aDistanceArray,
			      int *aSequence = NULL);


    /**
     * Gets the last received Odometry Pose. The CPose class contains the following members:
     *
     * double x;        - X position in Meters
     * double y;        - Y position in Meters
     * double heading;  - Heading in radians, CCW positive.
     *
     *
     * @param[out] aPosition    Last known position.
     * @param[out] aTimestamp   Pointer to timestamp. If this is non-NULL, it
     *                          will be filled with timestamp of last position
     *                          measure.
     *
     * @return                  true = Command succeeded
     *                          false = command failed!
     */
    bool ASRobo_GetPose(ASRobo::Core::CPose &aPosition,
			gim::time *aTimestamp);


    /**
     * Set the current Odometry Position to given value.
     *
     * @param[in] aPosition
     *                          double x;        - X position in Meters
     *                          double y;        - Y position in Meters
     *                          double heading;  - Heading in radians, CCW positive.
     * @return                  true = Command succeeded
     *                          false = command failed!
     *
     */
    bool ASRobo_SetPose(const ASRobo::Core::CPose &aPosition);


    /**
     * Set MCL Localization to 'aLoc'.
     * @param[in] aLoc          Localization position to set. Contains the following
     *                          members:
     *                          float x;              ///< X position (m)
     *	                        float y;              ///< Y position (m)
     *	                        float heading;        ///< Heading (rad)
     *                          float var_x;          ///< X position variance
     *                          float var_y;          ///< Y position variance
     *                          float var_heading;    ///< Heading variance
     *                          float probability;    ///< Position probability (0-1.00)
     * @return                  true = Command succeeded
     *                          false = command failed!
     */
    bool ASRobo_MCL_SetLocalization2D(const ASRobo::Drivers::Devices::Localize::TLocalization2D &aLoc);


    /**
     * Set current MCL Localization to 'aLoc'.
     * @param[out] aLoc         Store the current MCL position here.
     *                          Contains the following members:
     *                          float x;              ///< X position (m)
     *	                        float y;              ///< Y position (m)
     *	                        float heading;        ///< Heading (rad)
     *                          float var_x;          ///< X position variance
     *                          float var_y;          ///< Y position variance
     *                          float var_heading;    ///< Heading variance
     *                          float probability;    ///< Position probability (0-1.00)
     * @param[out] aTimestamp   Pointer to timestamp. If this is non-NULL, it
     *                          will be filled with timestamp of last position
     *                          measure.
     * @return                  true = Command succeeded
     *                          false = command failed!
     */
    bool ASRobo_MCL_GetLocalization2D(ASRobo::Drivers::Devices::Localize::TLocalization2D &aLoc,
                                      gim::time *aTimestamp);


    /**
     * Call Initialize for MCL
     *
     * @return                  true = Command succeeded
     *                          false = command failed!
     */
    bool ASRobo_MCL_Initialize(void);


    /**
     * Sets IO pin to state.
     * There are a few predefined constant values:
     *
     * ASROBO_IOPIN_SOLENOID (== 25)
     * ASROBO_IOPIN_LIGHT    (== 30)
     * ASROBO_IOPIN_LASER    (== 31)
     *
     *
     * @param[in] aPin         Pin number to set
     * @param[in] aState       State to set (true = HIGH, false = LOW)
     * @return                 true = Command succeeded
     *                         false = command failed!
     */
    bool ASRobo_IO_SetPin(const int aPin, const bool aState);


    /**
     * Get the state of the IO boards pin number 'aPin'
     * There are a few predefined constant values:
     *
     * ASROBO_IOPIN_SOLENOID (== 25)
     * ASROBO_IOPIN_LIGHT    (== 30)
     * ASROBO_IOPIN_LASER    (== 31)
     *
     * @param[in] aPin         Requested pin number.
     * @return                 true = The pin state is HIGH
     *                         false = The pin state is LOW (or command failed :I)
     */
    bool ASRobo_IO_GetPin(const int aPin);
		
    
    /**
     * Sets the state of the Secondary IO board (in front of the robot)
     * If you dont know the location of the Zero pin, ask the assistant
     * @param[in] aPin         Pin number to set Range must be (0...7)
     * @param[in] aState       State to set (true = HIGH, false = LOW)
     * @return                 true = Command succeeded
     *                         false = command failed!
     */
    bool ASRobo_ESC_IO_SetPin(const int aPin, const bool aState);
    
    
    /**
     * Get the state of the Secondary IO board (in front of the robot)
     * If you dont know the location of the Zero pin, ask the assistant
     * @param[in] aPin         Requested pin number. Range must be (0...7)!
     * @return                 true = The pin state is HIGH
     *                         false = The pin state is LOW (or command failed :I)
     */
    bool ASRobo_ESC_IO_GetPin(const int aPin);
    

    /**************************************************************************
     *************** ImageClient Interface ************************************
     *************************************************************************/

    /**
     * Read imagedata - most current data.
     *
     * @param[in] aCameraNumber Number of camera, Zero is the first camera.
     * @param[out] aImgPtr      Pointer to TBitmap structure. This structure
     *                          will be overwritten on each call. When the
     *                          object is no longer required,
     *                          call 'free_data()' for the object.
     *
     * @return                  true = Command succeeded
     *                          false = command failed!
     */
    bool IC_GetImage(const int aCameraNumber, TBitmap *aImgPtr);


    /**************************************************************************
     *************** GIMnet Interface *****************************************
     *************************************************************************/


    /**
     * Sends a freeform message over the GIMnet to recipient 'aReceiver'.
     * This message can be catched with 'GIMnet_ReceiveMessage'. You can
     * use this command to communicate between two j2b2GIMICtrl clients.
     * You decide yourself what kind of protocol you want to use - the message
     * is delivered to other party 'as-is'.
     *
     * @param[in] aReceiver     GIMnet name of the receiver.
     * @param[in] aDataPtr      Pointer to data to send
     * @param[in] aDataSize     Size of data in bytes to send (The buffer
     *                          pointed to by 'aDataPtr' must be atleas this
     *                          size)
     * @return                  true =  Command succeeded
     *                          false = command failed!
     */
    bool GIMnet_SendMessage(const std::string &aReceiver,
			    const void *aDataPtr,
			    const int aDataSize);

    /**
     * Receives a freeform message from the GIMnet. This only receives messages
     * previously sent with 'GIMnet_SendMessage'. If the call returns
     * succesfully, you can examine the data with following class member
     * functions:
     *
     *  int getSenderId() const;
     *  int getMessageFlags() const;
     *  int getDatalength() const;
     *  char* getData() const;
     *
     * See more information from header 'gimimessage.h', located under
     * ./j2b2GIMICtrl/GIMI/
     *
     * @param[out] aMsg         Message reference to fill with data.
     * @param[in] aTimeout_ms   Maximum time to wait for message. If this
     *                          time is reached, 'false' is returned. Set to
     *                          -1 to wait forever.
     * @return                  true = Message succesfully received
     *                          false = Timeout or other Error.

     */
    bool GIMnet_ReceiveMessage(gimi::GIMIMessage &aMsg,
			       const int aTimeout_ms);

    /**************************************************************************
     **************************************************************************
     **************************************************************************
     **************************************************************************
     ******* No user serviceable parts inside *********************************
     **************************************************************************
     **************************************************************************
     **************************************************************************
     **************************************************************************
     *************************************************************************/
  private:

    /**
     * Lock internal Mutex
     */
    void Lock(void);

    /**
     * Unlock internal Mutex.
     */
    void Unlock(void);

    ownMutexHandle iMutex;

    gimi::GIMI *iGIMInt;  ///< Pointer to GIMI interface

    /// ASRobo Pointers
    CASRoboClient *iASRoboInt; ///< Pointer to ASRobo interface
    CPositionClient *iASRoboInt_Position; ///< Pointer to ASRobo Position int
    CDrive2DClient *iASRoboInt_Drive2D;
    CRangingClient *iASRoboInt_Ranging_Laser; ///< Pointer to ASRobo Ranging int
    CRangingClient *iASRoboInt_Ranging_Bumper; ///< Pointer to ASRobo Ranging int
    CLocalizeClient *iASRoboInt_Localize_MCL;
    //CLocalizeClient *iASRoboInt_Localize_LaserOdometry;
    CIOClient *iASRoboInt_IO;
		CIOClient *iASRoboInt_ESCIO;

    /// ASRobo misc data
    std::string iASRoboServiceName; ///< Name of ASRobo service provider
    double iLastSpeed;
    double iLastAngularSpeed;
    double iLastAcceleration;

    /// PTU
    std::string iPTUServiceName; ///< Name of PTU serviceprovider
    int iLastPanRaw;
    int iLastTiltRaw;


    /// ImageClient Pointers
    std::string iICServiceName; ///< Name of IC serviceprovider
    CImageClient *iICInt; ///< Pointer to ImageClient interface

    /// Delayed link support
    int iLinkStaticDelay; ///< Static delay of the link per direction


    /// Private copy & assignment operators. This prevents copying this class.
    CJ2B2GIMICtrl(const CJ2B2GIMICtrl&)
      : iMutex(ownMutexHandleInitializer),
	iGIMInt(NULL),
	iASRoboInt(NULL),
	iASRoboInt_Position(NULL),
	iASRoboInt_Drive2D(NULL),
	iASRoboInt_Ranging_Laser(NULL),
	iASRoboInt_Ranging_Bumper(NULL),
	iASRoboInt_Localize_MCL(NULL),
	iASRoboInt_IO(NULL),
	iASRoboInt_ESCIO(NULL),
	iASRoboServiceName(),
	iLastSpeed(0),
	iLastAngularSpeed(0),
	iLastAcceleration(0),
	iPTUServiceName(),
	iLastPanRaw(0xFFFF),
	iLastTiltRaw(0xFFFF),
	iICServiceName(),
	iICInt(NULL),
	iLinkStaticDelay(0) {}
    CJ2B2GIMICtrl &operator=(const CJ2B2GIMICtrl &) { return *this; }

  };
}
#endif // _J2B2GIMICTRL_HPP_
